import importlib as imp
import tifffile,os,glob,pickle5,subprocess
import numpy as np,pandas as pd
from typing import TYPE_CHECKING, Dict, Iterable, Optional, Sequence, Union
from utils import displayStandards as dsp   #;imp.reload(dsp)
from utils import glob_colors as colors     #;imp.reload(colors)
from utils import handler3D as h3D          #;imp.reload(h3D)
from scipy.integrate import trapz
from scipy.stats import linregress
from . import utilities as ut               #;imp.reload(ut)

class Rocking:
    def __init__(self,Simu:object,
            uvw:list,tag:str,path:str,
            Sargs:dict):
        """ simulate rocking curve

        Parameters
        ----------
        Simu
            Simulator object
        uvw
            list of orientation vectors
        tag
            a tag identifying the sweep
        Sargs
            Simulator constructor arguments
        """
        self.path = path
        self.tag  = tag
        self.uvw  = uvw
        self.Sargs = Sargs
        self.df = ut.sweep_var(Simu,params='u',vals=uvw,tag=tag,path=path,**Sargs)
        self.n_simus=uvw.shape[0]
        ts            = np.arange(self.n_simus)
        self.ts       = ts
        self.df['ts'] = ts
        self.Iz_dyn = {}
        self.Iz_kin = {}
        self._build_index()
        self.save(v=1)

    ###########################################################################
    #### compute
    ###########################################################################
    def _build_index(self):
        frames = [ i for i,name in enumerate(self.df.index)]
        beams = np.unique(np.hstack([self.load(i).df_G.index.values for i,name in enumerate(self.df.index)]))
        self.nbeams = beams.size
        print('...building index...')
        self.beams = pd.DataFrame(index=beams,columns=['Frame','Sw'])
        for h in beams:
            self.beams.loc[h,'Frame']=[]
            self.beams.loc[h,'Sw']=[]
        for i,f in enumerate(self.df.index):
            for h,c in self.load(i).df_G.iterrows():
                self.beams.loc[h,'Frame'].append(i)
                self.beams.loc[h,'Sw'   ].append(c.Sw.real)
        self.df['nbeams']=[self.load(i).df_G.index.shape[0] for i,f in enumerate(self.df.index)]
        self.save(v=1)

    def reset_int(self):
        self.Iz_dyn,self.Iz_kin = {},{}

    def _integrate_rocking(self,refl=[],new=0):
        if not new:
            # refl,nbs = self.get_beams(cond=cond,refl=refl)
            nbs=len(refl)
            refl  = [h for h in refl if not h in self.Iz_dyn.keys()] #;print(hkl)

        hkl = [eval(h) for h in refl]
        z,nzs = self._get_z()
        nbs,nts = len(hkl),self.ts.size
        if nbs:
            Iz_dyns  = dict(zip(refl, [np.zeros((nts,nzs)) for h in refl] ))
            Iz_kins  = dict(zip(refl, [np.zeros((nts,nzs)) for h in refl] ))
            for i in range(nts):
                sim_obj = self.load(i)
                idx = sim_obj.get_beam(refl=refl,cond='',index=True)
                if idx:
                    hkl0 = sim_obj.df_G.iloc[idx].index #[str(tuple(h)) for h in sim_obj.get_hkl()[idx]]
                    for idB,hkl_0 in zip(idx,hkl0):
                        Iz_dyns[hkl_0][i,:] = sim_obj.Iz[idB,:]
                        Iz_kins[hkl_0][i,:] = sim_obj.Iz_kin[idB,:]

            Iz_dyn  = dict()
            Iz_kin  = dict()
            for h in refl:
                df_b = self.beams.loc[h]
                if len(df_b.Sw)>1:
                    s = np.sign(df_b.Sw[1]-df_b.Sw[0])
                    Iz_dyn[h]=[s*trapz(Iz_dyns[h][df_b.Frame,iz],df_b.Sw) for iz in range(nzs)]
                    Iz_kin[h]=[s*trapz(Iz_kins[h][df_b.Frame,iz],df_b.Sw) for iz in range(nzs)]
                else:
                    Iz_dyn[h] = [Iz_dyns[h][df_b.Frame,iz] for iz in range(nzs)]
                    Iz_kin[h] = [Iz_kins[h][df_b.Frame,iz] for iz in range(nzs)]
            self.Iz_dyn.update(Iz_dyn)
            self.Iz_kin.update(Iz_kin)
            self.save()
            print(colors.green+'rock.Iz updated'+colors.black)

    def get_rocking(self,iZs:Optional[slice]=-1,
        zs:Optional[Iterable[float]]=None,
        refl:Sequence[tuple]=[],cond:str='',opts:str='',n=0):
        """Get intensities at required beams and thicknesses

        Parameters
        ----------
        zs
            selected thicknesses
        iZs
            slice into thicknesses
        refl
            reflections to get
        cond
            condition to apply to selecte reflections
        opts
            F(include friedel), O(include central beam)

        Returns
        -------
            z,dict
                z,{beam:I(z)}
        """
        iZs,nzs  = self._get_iZs(iZs,zs)        #;print(iZs)
        z0 = self.load(0).z.copy()[iZs][-1]
        if not self.load(0).thick==z0:
            print('setting thickness to %dA' %z0)
            self.do('_set_I',v=0,iZ=iZs[-1])

        refl,nbs = self.get_beams(cond=cond,refl=refl,opts=opts)  #;print(refl)
        nts = self.ts.size
        I = {}
        # for h in refl : I[str(h)]=np.nan*np.ones((nts,nzs))
        for h in refl : I[str(h)]=np.zeros((nts,nzs))

        print("gathering the intensities")
        for i in range(nts):
            # print(colors.red,i,colors.black)
            sim_obj = self.load(i)
            hkl0  = sim_obj.get_beam(refl=refl,index=False)
            if hkl0:
                idx  = sim_obj.get_beam(refl=refl,index=True)
                # hkl0 = [str(tuple(h)) for h in sim_obj.get_hkl()[idx]]
                for idB,hkl_0 in zip(idx,hkl0):
                    I[hkl_0][i,:] = np.array(sim_obj.Iz[idB,iZs])
        if n and nbs>n:
            print('keeping only %d strongest beams' %n)
            df_Imax = pd.DataFrame.from_dict({h:Ib[:,0].max() for h,Ib in I.items()},orient='index',columns=['I'])
            hkls = df_Imax.sort_values('I',ascending=False)[:n].index
            I = {h:I[h] for h in hkls}

        z = self.load(0).z.copy()[iZs]
        return z,I

    def integrate(self,thick=None):
        if not thick:
            return self._integrate_all()
        else:
            return self._integrate_thick(thick)

    def _integrate_thick(self,thick):
        b0 = self.load(0)
        if not b0.thick==thick:
            self.do('set_thickness',thick=thick)
        df_int = pd.DataFrame(0,index=self.beams.index,columns=['I'])
        for i in range(self.n_simus):
            b0 = self.load(i)
            df_int.loc[b0.df_G.index,'I'] += b0.df_G.I
        # self.save()
        return df_int

    def _integrate_all(self):
        b0 = self.load(0)
        thicks=['%.1fA' %z for z in b0.z ]
        self.z=b0.z
        self.df_int = pd.DataFrame(0,index=self.beams.index,columns=thicks)

        for i in range(self.n_simus):
            b0 = self.load(i)
            self.df_int.loc[b0.df_G.index] += b0.Iz
        self.save()
        # print(self.z.shape,self.df_int.values.shape)
        return self.df_int

    def Rfactor(self,df_exp):
        refl = self.df_int.loc[self.df_int.index.isin(df_exp.index)].index
        I_exp = df_exp.loc[refl,'I'].values             #;print(I_exp)
        prescale=I_exp.mean() 

        z=self.df_int.columns
        self.R = pd.DataFrame(index=z,columns=['scale','r_value','Rfac'])
        Isum = I_exp.sum()
        for z0 in z :
            I_sim = self.df_int.loc[refl,z0].values*prescale     #;print(I_sim)
            scale, intercept, r_value, p_value, std_err = linregress(I_sim, I_exp)
            Rfac = abs(I_sim*scale - I_exp).sum()/Isum
            self.R.loc[z0] = [scale,r_value,Rfac]

        self.save()
        return self.R
    ###########################################################################
    #### Display
    ###########################################################################
    # def QQplot(self,zs=None,iZs=10,refl=[],cond='',
    #     int_opt=True,cmap='Spectral',**kwargs):
    #     if int_opt:self.integrate_rocking(refl,cond)
    #     iZs,nzs  = self._get_iZs(iZs,zs)    #;print(iZs)
    #     z  = self.load(0).z.copy()[iZs]
    #
    #     refl = self.get_beams(refl=refl,cond=cond)
    #     refl = [str(tuple(h)) for h in refl]
    #     Iz_dyn = np.array([self.Iz_dyn[h].copy()[iZs] for h in refl])
    #     Iz_kin = np.array([self.Iz_kin[h].copy()[iZs] for h in refl])
    #     # print(Iz_dyn.shape)
    #     iB = np.argsort(np.sum(Iz_dyn,axis=1))[-1]
    #     Iz_dyn/= Iz_dyn[iB,:]
    #     Iz_kin/= Iz_kin[iB,:]
    #     cs = dsp.getCs(cmap,nzs) #; print(len(cs),Iz_dyn.shape,Iz_kin.shape)
    #
    #     plts=[[I_kin,I_dyn,[cs[i],'o'],r'$z=%d \AA$' %z0] for i,(z0,I_dyn,I_kin) in enumerate(zip(z,Iz_dyn.T,Iz_kin.T))]
    #     plts+=[ [[0,1],[0,1],[(0.5,)*3,'--'],''] ]
    #     dsp.stddisp(plts,labs=['$I_{kin}$','$I_{dyn}$'],sargs={'alpha':0.5},**kwargs)

    def plot_integrated(self,refl,new:bool=False,cm='Spectral',**kwargs):
        """plot the integrated rocking curves for selected beams as function of z

        Parameters
        ----------
        refl
            beams to consider
        new
            update integration
        """
        self._integrate_rocking(refl=refl,new=new)
        nbs = len(refl)
        z = self.load(0).z
        Iz = np.array([self.Iz_dyn[h] for i,h in enumerate(refl)])

        cs = dsp.getCs(cm,nbs)
        plts = [[z,Iz[i],cs[i],'%s' %h] for i,h in enumerate(refl)]
        fig,ax=dsp.stddisp(plts,labs=[r'$z(\AA)$','$I_{int}$'],**kwargs)
        vw=Rock_viewer(self,fig,ax,z,Iz,refl)
        return fig,ax


    def plot_rocking(self,cmap='viridis',x:str='Sw',
        cond='',refl=[],opts:str='',iZs=-1,zs=None,n:int=0,
        **kwargs):
        """plot rocking curve for set of selected beams at thickness zs

        Parameters
        ----------
        cond,refl,iZs,zs,opts
            select beams and thicknesses (see :meth:`~Rocking.get_rocking`)
        x
            to display on x axis
        kwargs
            args to pass to dsp.stddisp
        """
        z,I = self.get_rocking(cond=cond,refl=refl,opts=opts,zs=zs,iZs=iZs,n=n)
        refl,plts = list(I.keys()),[]           #;print(refl)
        xlab = {'frame':'frame','Sw':r'$S_w(\AA^{-1})$','theta':r'$\theta(deg)$'}[x]

        print('gathering plots')
        nbs,nzs = len(refl),z.size
        if nbs>=nzs:
            cs,ms = dsp.getCs(cmap,nbs), dsp.markers
            legElt = { '%s' %refl0:[cs[i],'-'] for i,refl0 in enumerate(refl)}
            for iz,zi in enumerate(z):
                legElt.update({'$z=%d A$' %(zi):['k',ms[iz]+'-']})
                for i,refl0 in enumerate(refl):
                    df_b=self.beams.loc[refl0]
                    plts += [[df_b.Sw,I[refl0][df_b.Frame,iz],[cs[i],ms[iz]+'-'],'']]
        else:
            # rocking for different thicknesses
            cs,ms = dsp.getCs(cmap,nzs),  dsp.markers
            legElt = { '%s' %refl0:['k','-'+ms[i]] for i,refl0 in enumerate(refl)}
            # self.get_frames(hkl,iTs=slice(0,None))
            for i,refl0 in enumerate(refl):
                for iz,zi in enumerate(z):
                    df_b=self.beams.loc[refl0]
                    plts += [[df_b.Sw,I[refl0][df_b.Frame,iz],[cs[iz],ms[i]+'-'],'']]
            legElt.update({'$z=%d A$' %(zi):[cs[iz],'-'] for iz,zi in enumerate(z) })

        # print('displaying')
        return dsp.stddisp(plts,labs=[xlab,'$I$'],legElt=legElt,**kwargs)

    def Sw_vs_theta(self,refl=[[0,0,0]],cond='',thick=None,fz=abs,opts='',
        iTs=slice(0,None),ts=None,
        cm='Spectral',figname='',**kwargs):
        """Displays Sw and I for a range of angle simulations at given thickness

        Parameters
        ----------
        refl,cond
            selected reflections
        thick
            thickness
        fz
            functor for Sw
        Iopt
            plot I
        """
        Iopt = 'I' in opts
        iTs,nts = self._get_iTs(iTs,ts)
        xlab,ts = r'$\theta$',self.ts.copy()[iTs]
        if 'f' in opts:
            xlab,ts = 'frame',np.arange(1,self.ts.size+1)[iTs]       #;print(iTs,ts)

        # dsp.stddisp([[r.Frame,r.Sw,[c,'-o'],h] for c,(h,r) in zip(dsp.getCs('jet',len(refl)),rock.beams.loc[refl].iterrows()) ] )

        if Iopt:
            if thick:
                if not self.load(0).thick==thick:
                    # self.do('_set_excitation_errors',Smax=0.025)
                    # self.do('solve',thick=thick,Smax=0.025)
                    self.do('set_thickness',thick=thick)
            else:
                thick = self.load(0).thick
        refl,nbs = self.get_beams(cond,refl)                        #;print(refl)

        Sw = pd.DataFrame(np.ones((nts,nbs)),columns=[str(h) for h in refl])
        if Iopt:I  = pd.DataFrame(np.zeros((nts,nbs)),columns=[str(h) for h in refl])
        for i,name in enumerate(self.df.index[iTs]):
            b = self.load(i) #;print(i)
            hkl0 = b.get_beam(refl=refl,cond=cond,opt=0)
            # hkl0 = [str(tuple(h)) for h in b.get_hkl()[idx]]
            Sw.loc[i,hkl0] = b.df_G.loc[hkl0,'Sw'].values
            if Iopt:I.loc[i,hkl0] = b.df_G.loc[hkl0,'I'].values

        #locate minimum excitation errors
        iSmin = np.argmin(abs(Sw).values.T,axis=1) #locate minimums
        Sw[Sw==1]=np.nan
        SwE = fz(Sw.values.T)

        if 'i' in opts:
            dsp.stddisp(im=[SwE],pOpt='im',labs=[xlab,'$beam$'],caxis=[fz(1e-2),fz(1e-6)],
                cmap='Reds',title='Excitation error',name=figname+'_Sw.svg',**kwargs)
            # print(refl)

        else:
            cs,txts = dsp.getCs(cm,nbs),[]
            plts = [[ts,Sw0,[cs[i],'-o'],''] for i,Sw0 in enumerate(SwE)]
            if 't' in opts:txts = [[ts[idx],SwE[i,idx],'%s' %str(h),cs[i]] for i,(h,idx) in enumerate(zip(refl,iSmin))]
            dsp.stddisp(plts,texts=txts,labs=[xlab,'$S_w$'],name=figname+'_Sw.svg',**kwargs)

        if Iopt:
            IE = I.values.T
            if 'i' in opts:
                dsp.stddisp(im=[IE],pOpt='im',labs=[xlab,'$beam$'],caxis=[0,0.2],
                    cmap='YlGnBu',title='Intensity',name=figname+'_Sw.svg',**kwargs)
            else:
                if 'w' in opts:
                    xlab='$S_w$'
                    plts = [[Sw0,I0,[cs[i],'-o'],''] for i,(Sw0,I0) in enumerate(zip(SwE,IE))]
                    txts = [[0,IE[i,idx],'%s' %str(h),cs[i]] for i,(h,idx) in enumerate(zip(refl,iSmin))]
                else:
                    plts = [[ts,I0,[cs[i],'-o'],''] for i,I0 in enumerate(IE)]
                    txts = [[ts[idx],IE[i,idx],'%s' %str(h),cs[i]] for i,(h,idx) in enumerate(zip(refl,iSmin))]
                dsp.stddisp(plts,texts=txts,labs=[xlab,'$I$'],
                    title=r'thickness=$%d A$' %thick,name=figname+'_I.svg',**kwargs)
            # return IE
        # return Sw

    ###########################################################################
    #### gets
    ###########################################################################
    def get_frames(self,hkl:str,cols:Sequence[str]=['Sw','I']):
        """get frames in which beam hkl appears

        Parameters
        ---------
        hlk
            single reflection to gather information about
        cols
            info to fetch

        Returns
        -------
        DataFrame with info for that reflection
        """
        frames = [ i for i,name in enumerate(self.df.index)
            if hkl in self.load(i).df_G.index]

        df = pd.DataFrame(frames,columns=['frame'])
        df[cols]=0

        for i,f in enumerate(frames):
            vals = self.load(f).df_G.loc[hkl,cols].values #;print(f,vals)#df.iloc[i,cols])
            df.loc[i,cols] = vals.real
        return df

    def get_beams(self,cond='',refl=[],opts='',n=None):
        if cond:
            refl = []
            for i,name in enumerate(self.df.index):
                b = self.load(i)
                refl += b.get_beam(cond=cond,index=False)
        refl = np.unique(refl)           #;print(refl)
        if not isinstance(refl[0],str):
            refl=[str(tuple(h)) for h in refl]
        if not 'F' in opts:
            refl=ut.remove_friedel_pairs(refl)
        if 'O' in opts:
            refl=np.hstack([refl,str((0,0,0))])

        if n and len(refl)>n:refl=refl[:n]
        nbs = len(refl)#.size;print(nbs,refl)
        print('total number of beams:%d' %nbs)

        return refl,nbs

    def _get_iTs(self,iTs,ts):
        t = self.ts
        if isinstance(ts,float) or isinstance(ts,int):ts = [ts]
        if isinstance(ts,list) or isinstance(ts,np.ndarray):
            iTs = [np.argmin(abs(t-t0)) for t0 in ts]
        if isinstance(iTs,int):iTs=[iTs]
        if not type(iTs) in [list,np.ndarray,slice]:iTs = slice(0,None,1)
        nts = t[iTs].size
        return iTs,nts

    def _get_iZs(self,iZs,zs):
        z = self.load(0).z
        if isinstance(zs,float) or isinstance(zs,int):zs = [zs]
        if isinstance(zs,list) or isinstance(zs,np.ndarray):
            iZs = [np.argmin(abs(z-z0)) for z0 in zs]
        if isinstance(iZs,int):iZs=[iZs]
        nzs = z[iZs].size
        return iZs,nzs

    def _get_z(self):
        z = self.load(0).z
        nzs = z.size
        return z,nzs

    def _get_ts(self,i,ts):
        if type(ts) in [float,int]:i=np.argmin(abs(self.ts-ts))
        return i,self.ts[i]

    ###########################################################################
    #### misc
    ###########################################################################
    def set_tag(self,tag):
        nts   = self.ts.size
        pad   = int(np.ceil(np.log10(nts)))
        names = [ '%s_%s%s' %(tag,'u',str(i).zfill(pad)) for i in range(nts)]
        self.df.index = names
        cmd ="cd %s; rename 's/%s/%s/' %s*.pkl df_%s.pkl rock_%s.pkl" %(self.path,self.tag,tag,self.tag,self.tag,self.tag)
        p = subprocess.Popen(cmd,shell=True);p.wait()
        self.tag = tag
        for i,name in enumerate(names):
            sim_obj = load_pkl(os.path.join(self.path,name+'.pkl'))
            sim_obj.set_name(name=name,path=self.path)
            sim_obj.save()
            self.df.loc[name,'pkl'] = sim_obj.get_pkl()
        df_file = os.path.join(self.path,'df_%s.pkl' %tag)
        self.df.to_pickle(df_file)
        print(colors.green+'Dataframe saved : '+colors.yellow+df_file+colors.black)
        self.save()

    def save(self,v=1):
        """save this object"""
        file = rock_name(self.path,self.tag)
        with open(file,'wb') as out :
            pickle5.dump(self, out, pickle5.HIGHEST_PROTOCOL)
        if v:print(colors.green+"object saved\n"+colors.yellow+file+colors.black)

    def load(self,i:int=0,ts:float=None):
        """load one of the simu

        Parameters
        ----------
        i
            index of the simu
        ts
            value
        """
        i,ts = self._get_ts(i,ts)
        file = self.df.iloc[i].pkl
        # file = os.path.join(self.path,os.path.basename(self.df.iloc[i].pkl))
        sim_obj = ut.load_pkl(file)
        return sim_obj

    def change_path(self,path):
        self.path=path
        self.df.pkl=[s.replace(os.path.dirname(s),self.path) for s in self.df.pkl]
        self.save()
        for i in range(self.df.shape[0]):
            obj = self.load(i)
            obj.path=path
            obj.save()

    def do(self,f,verbose=True,**args):
        """ apply function to all simus
        """
        for i in range(self.df.shape[0]):
            obj = self.load(i)
            obj.__getattribute__(f)(**args)
            obj.save(v=verbose)


def rock_name(path,tag):
    return os.path.join(path,'rock_%s.pkl' %(tag))

def load_rock(path,tag):
    return load_pkl(file=exp.rock_name(path,tag))

class Rock_viewer:
    def __init__(self,rock,fig,ax,z,Iz,refl):
        self.rock=rock
        self.fig = fig
        self.ax = ax
        self.z  = z
        self.Iz = Iz
        self.refl=refl
        cid = self.fig.canvas.mpl_connect('key_press_event', self)

    def __call__(self, event):
        # print(event.key)
        keys={
            'single':'enter' ,
            'thicks':'z'     ,
        }
        if event.key == keys['single']:
            z,I=dsp.plt.ginput(n=1)[0]
            iz = abs(self.z-z).argmin()
            ih = abs(I-self.Iz[:,iz]).argmin()
            print(iz,ih)
            refl,iZs = [self.refl[ih]],[iz]

        elif event.key == keys['thicks']:
            dats=dsp.plt.ginput(n=-1)
            iZs = [abs(self.z-z).argmin() for z,I in dats]
            I,z = dats[0]
            ih  = abs(I-self.Iz[:,iZs[0]]).argmin()
            refl,iZs = [self.refl[ih]],iZs

        if event.key in keys.values():
            self.rock.plot_rocking(refl=refl,iZs=iZs)



from utils import*
import matplotlib.tri as mtri
from scipy.spatial import Delaunay

def nodesTri(nDeg) : 
    h = 1/(nDeg-1)
    # Equally-space nodes 
    nodes = np.array([[],[]])
    for j in range(nDeg) : 
        i = np.arange(nDeg-j).T
        aux = j*np.ones(i.size)
        xy = np.array([i*h,aux*h]); #print(xy)
        nodes =  np.concatenate((nodes,xy),axis=1)
    nodes = 2*nodes.T-1
    return nodes


def test():
    nDeg = 2
    nodes = nodesTri(nDeg)
    #use
    x,y = nodes.T
    x0,y0=-0.05,-0.15
    f = (x-x0)**2 + (y-y0)**2
    fig, ax = plt.subplots()
    triang = mtri.Triangulation(x, y, Delaunay(nodes).vertices)
    ax.tricontourf(triang, f,levels=20)
    ax.triplot(triang)
    ax.plot(x,y, 'o')
    fig.tight_layout()
    plt.show()
    
if '__main__'==__name__:
    test()


ssh_hosts = {
    'local_london':'BG-X550',
    'tarik-CCP4stfc':'tarik-CCP4','tarik-CCP4home':'tarik-CCP4',
    'brno':'brno',
    'badb':'badb',
    'badb.rc-harwell.ac.uk':'badb',
    'stfc-cloud':'stfc-cloud',
    'ms01':'ms01',
}

temsim_hosts={
    'BG-X550'   :'/home/ronan/Documents/git/ccp4/src/electron-diffraction/multislice/bin/',
    'tarik-CCP4':'/home/tarik/Documents/git/ccp4/src/electron-diffraction/multislice/bin/',
    'brno'      :'/home/tarik/Documents/git/ccp4/src/electron-diffraction/multislice/bin/',
    'test'      :'/home/lii26466/Documents/git/ccp4/src/electron-diffraction/multislice/bin/',
    'ms01'      :'/home/lii26466/Documents/git/ccp4/src/electron-diffraction/multislice/bin/',
    'badb'      :'/home/lii26466/Documents/git/ccp4/src/electron-diffraction/multislice/bin/',
    'badb.rc-harwell.ac.uk':'/home/lii26466/Documents/git/ccp4/src/electron-diffraction/multislice/bin/',
}
path_hosts={
    'BG-X550'   :'/home/ronan/Documents/git/ccp4/src/electron-diffraction/multislice/dat/',
    'tarik-CCP4':'/home/tarik/Documents/git/ccp4/src/electron-diffraction/multislice/dat/',
    'brno'      :'/home/tarik/Documents/git/ccp4/src/electron-diffraction/multislice/dat/',
    'test'      :'/home/lii26466/Documents/git/ccp4/src/electron-diffraction/multislice/dat/',
    'ms01'      :'/home/lii26466/Documents/git/ccp4/src/electron-diffraction/multislice/dat/',
    'badb'      :'/data3/lii26466/multislice/',
    'badb.rc-harwell.ac.uk' :'/data3/lii26466/multislice/',
}



import glob,os,re
import tifffile,mrcfile,cbf,numpy as np
from utils import glob_colors as colors
from utils import displayStandards as dsp
import cv2

nb_colors=100
cmaps = ['hot','viridis','Spectral','Greys']

cs  = {cmap : (np.array(dsp.getCs(cmap,nb_colors))*255).astype(np.uint8)
    for cmap in cmaps}

def read(file):
    fmt=file.split('.')[-1]#;print(fmt)
    return img_readers[fmt](file)

def tiff_reader(tiff_file)  :
    with open(tiff_file,'rb') as f:
        I = tifffile.imread(f)
    return I

def cbf_reader(cbf_file):
    content = cbf.read(cbf_file)
    numpy_array_with_data = content.data
    # header_metadata = content.metadata
    # print(colors.blue,header_metadata,colors.black)
    return numpy_array_with_data

def mrc_reader(mrc_file):
    with mrcfile.open(mrc_file) as mrc:
        return mrc.data

def smv_reader(file,head=False):
    with open(file, "rb") as f:
        buf= f.read()

    header = [l.strip('\n;') for l in buf[:512].decode().split('\n')]
    header = {k: v for k, v in map(lambda s: s.split("="), header[1:-3])}
    if head: print(header)
    h_bytes = int(header['HEADER_BYTES'])
    assert(header['TYPE']=='unsigned_short')
    nx,ny=int(header['SIZE1']),int(header['SIZE2'])
    img = np.reshape(
        np.frombuffer(buf[h_bytes:], dtype=np.uint16),
        (nx,ny))
    return img


img_readers = {
    'mrc' : mrc_reader,
    'tiff': tiff_reader,
    'tif' : tiff_reader,
    'cbf' : cbf_reader,
    'smv' : smv_reader,
    'img' : smv_reader,
}
fmts = list(img_readers.keys())

def detect_frame(path):
    files = glob.glob(os.path.join(path,'*.*'))
    if not len(files):return
    #detect format
    fmt,i = '',0
    while fmt not in fmts and i<len(files):
        fmt = files[i].split('.')[-1]
        i+=1

    if fmt:
        frames    = np.sort(glob.glob(os.path.join(path,'*.%s' %fmt))).tolist()
        nb_frames = len(frames)
        min_pad   = int(np.ceil(np.log10(nb_frames)))
        frame0 = os.path.basename(frames[0])
        frame1 = os.path.basename(frames[-1])
        #get the frame string which contains padding
        #if there are other occurences of numbers in the frame, take the one that changes from frame to frame
        min_frame_str = np.array(re.findall(r"[0-9]{%d,}" %(min_pad),frame0),dtype=str)
        max_frame_str = np.array(re.findall(r"[0-9]{%d,}" %(min_pad),frame1),dtype=str)

        if len(min_frame_str)>1:
            idx =~(min_frame_str==max_frame_str)
            min_frame_str=min_frame_str[idx]
            max_frame_str=max_frame_str[idx]
        min_frame_str = min_frame_str[0]
        max_frame_str = max_frame_str[0]

        prefix,suffix = frame0.replace('.%s' %fmt,'').split(min_frame_str)
        pad       = len(min_frame_str)
        min_frame = int(min_frame_str)
        max_frame = int(max_frame_str)

        # print('frames info : ', frame0,frame_str,pad,prefix,' ',min_frame)
        # max_frame    = int(os.path.basename(frames[-1]).replace('.'+fmt,'').replace(prefix,''))
        frames_dict = {
            'nb_frames' : nb_frames,
            'min_frame' : min_frame,
            'max_frame' : max_frame,
            'pad'       : pad,'fmt':fmt,
            'template'  : '%s%%s%s.%s' %(prefix,suffix,fmt),
        }
        # print(frames_dict)
    return frames_dict


def save_image(file,zmax=1000,cmap='hot',dir='',v=False):
    if v:print('reading...')
    im  = read(file)    #;print(im.shape)

    if v:print('processing...')
    idx = np.maximum(0,np.floor((np.minimum(im,zmax-1)/zmax)*nb_colors)).astype(np.int)

    fmt = file.split('.')[-1]
    out = os.path.join(dir,os.path.basename(file).replace(fmt,'jpg'))
    if v:print('writing...')
    success=cv2.imwrite(out, cs[cmap][idx,:])

    if success:
        print(colors.green,'file saved : ')
        print(colors.yellow,out,colors.black)
    else:
        print(colors.red,'saving failed',colors.black)



import importlib as imp
import easygui,os,glob,copy,pickle
from subprocess import Popen,PIPE
import numpy as np,matplotlib.pyplot as plt,pandas as pd
from utils import displayStandards as dsp
from utils import glob_colors as colors
# from multislice.pets import Pets
# from multislice.multislice import Multislice
# print('...bloch,multi...')
from blochwave.util import load_bloch #Bloch
from multislice import mupy_utils as mut    #;imp.reload(mut)
from multislice import postprocess as pp    #;imp.reload(pp)
from multislice import multislice as ms     #;imp.reload(ms)
from blochwave import bloch as bl           ;imp.reload(bl)
from blochwave import util as bl_ut         ;imp.reload(bl_ut)
from . import __version__
from . import pets as pt                    ;imp.reload(pt)
from . import display as EDdisp             ;imp.reload(EDdisp)
from . import gui_config as cfg             ;imp.reload(cfg)
from . import utilities as ut               ;imp.reload(ut)

# currently bugs with pandas 1.4
#pd.set_option('precision',3)
#pd.set_option('max_rows',100)
if plt.rcParams['keymap.save']:
    plt.rcParams['keymap.save']=[]
    plt.rcParams['keymap.grid']=[]
    plt.rcParams['keymap.xscale']=[]
    plt.rcParams['keymap.yscale']=[]
    # plt.rc('text', usetex=False)



class Gui:
    def __init__(self,config=None,help=True,pargs={},**kwargs):
        '''Dynamic beam viewer :
            - config : str - saved configuration
            - help : bool - show help at startup
            - kwargs : see init_args
        '''
        self.__version__=__version__
        # print('...setting up keys...')
        # self.set_keys()
        self.keymap = cfg.KeyBinder()
        if help:self.keymap.show_help_cli()

        self.path,self.cif_file,self.bloch,self.pets = None,None,None,None
        self.multi_path,self.tifpath,self.multi = None,None,None
        self.rpl,self.im,self.Icols,self.rock,self.dyn_opt = None,None,[],None,1
        self.cond = '(Sw<1e-3) & (Vga>1e-5)'
        self.nfigs,self.kargs,self.u = 0,{},[0,0,1]
        self.pattern_args = {'Iopt':'csngt','Imax':5e4,'gs':0.025,'rmax':25,'Nmax':512}
        self.multi_args = {'data':'','mulslice':False,'NxNy':512,'Nhk':0,'hk_pad':None,
            'slice_thick':1,'i_slice':10,'ssh':'','opt':'srf'}
        self.xyz_params = {'theta':0,'lat_params':[20,20,100],'pad':1.0,'rep':None}
        self.dsp_d = dict(zip(self.keymap.dsp_keys,np.zeros((len(self.keymap.dsp_keys)),dtype=bool) ))
        self.opt_d = dict(zip(self.keymap.opt_keys,np.zeros((len(self.keymap.opt_keys)),dtype=bool) ))
        self.show_Sw,self.Swtol=False,2
        self.m={'I':1e3,'Ig':1e2,'Vg':10,'Swl':self.Swtol}

        # print('...load parameters...')
        if config:self.load_config(config)
        else:self.init_args(**kwargs)
        self.set_dsp_d()

        self.cutoff_r,self.mag_r,self.dthick_r=self.cutoff,self.mag,self.dthick
        if 'tag' in self.multi_args.keys():
            self.multi_args['tail']=self.multi_args['tag']
            self.multi_args.pop('tag')

        self.load_objects()
        self.multi_args['tail'] = self.tag+self.frame_str()
        self.multi_args['name'] = self.multi_path

        self.set_mode(self.mode)
        if not (self.tifpath or self.multi or self.rock):self.set_mode('rotate')

        # print('...Complete initialization...')

        #print('... init figure ...')
        self.pargs=pargs
        self.fig,self.ax = dsp.stddisp(**pargs )
        cid = self.fig.canvas.mpl_connect('key_press_event', self)
        self.call = self.call_update

        # print('... update and show ...')
        if self.multi_path:
            self.xyz_params['xyz']  = self.get_xyz()
        self.xyz_params['file'] = self.cif_file
        self.xyz_params['n']    = self.u
        if not self.thick : self.thick=100
        if not self.rock:
            self.update(fsolve=1)
            self.update_thickness()
        self.show()
        # plt.show()


    def init_args(self,path=None,cif_file=None,bloch=None,rock=None,pets=None,multi=None,tag='',
        orient=[0,0,5,5],u=None,Smax=0.01,Nmax=5,thick=None,dthick=5,thicks=(0,1000,1000),
        xylims=1.5,cutoff=50,dcutoff=10,mag=500,cmap=None,cmap_beams='jet',
        frame=1,incrate=1,drot=1,drotS=1,pets_opts='Pr',rot=0,rotS=0,dyn_opt=1,
        init_opts='R'):
        # F='Sw',kargs={},**kwargs):
        ''' configuration Initializer
    #### path setup (at least one of these need to be defined)
        - path      : str - path to the main folder
        - cif_file  : str - full path to .cif file (will be automatically fetched in path if not defined)
        - rock      : Rocking object or str for a rocking object
        - bloch     : Bloch object or str (will be automatically created if not defined)
        - multi     : str - path to multislice simulations (default multi)
        - pets      : Pets object or str to .pts file -
    #### Simulation setup
        - orient : [theta,phi,dtheta,dphi] in degrees
        - u      : list3 - set beam direction(takes preference over orient)
        - Smax,Nmax,thick,F : see blochwave.Bloch
        - init_opts: str -
            - h(show help at start) R(rotate mode) B(save Bloch)
            - k(show_hkl) i(show_i) z(show_z) u(show_u) v(show u in zone axis reciprocal space)
        - cutoff : brightness cutoff
        - mag    : circle magnification
        '''
        #generic
        self.cutoff  = cutoff
        self.dcutoff = dcutoff
        self.mag     = mag
        self.xylims  = xylims
        self.Smax    = Smax
        self.Nmax    = Nmax
        self.thick   = thick
        self.dthick  = dthick
        self.cmap    = cmap
        self.rot ,self.drot  = rot, drot
        self.rotS,self.drotS = rotS, drotS

        #rotate related
        self.theta,self.phi,self.dtheta,self.dphi = orient
        self.dtheta_dphi = [self.dtheta,self.dphi]
        #frames related
        self.frame   = frame
        self.incrate = incrate
        #bloch related
        self.cmap_beams = cmap_beams
        self.thicks  = thicks #; self.F,self.fopts = F,self.df_F.loc[self.df_F.F==F].iloc[0].fopt
        #pets related
        self.pets_opts = pets_opts
        self.dyn_opt = dyn_opt
        #multislice related
        self.tag  = tag

        #init_opts
        self.mode       = ['frames','rotate']['R' in init_opts]
        self.show_hkl   = 'k' in init_opts
        self.show_i     = 'i' in init_opts
        self.show_Ig    = 'I' in init_opts
        self.show_Vg    = 'V' in init_opts
        self.show_Sw    = 'W' in init_opts
        self.show_z     = 'z' in init_opts
        self.show_u     = 'u' in init_opts
        self.show_v     = 'v' in init_opts
        self.save_bloch = 'B' in init_opts
        self.hold_mode  = ' ' in init_opts

        # print('...init path...')
        self.set_theta_phi_from_u(u)
        self.init_path(path,cif_file,bloch,rock,pets,multi)

    def load_config(self,config):
        with open(config,'rb') as f:d_config = pickle.load(f)
        for k,v in d_config.items(): self.__dict__[k] = v

    def init_path(self,path,cif_file,bloch,rock,pets,multi):
        if all([type(o)==type(None) for o in [path,cif_file,bloch,pets] ]):
            args = ['path','cif_file','bloch','pets']
            raise Exception('at least one of these must be defined : ',args)

        if type(path)==str:self.path = path
        if type(cif_file)==str:self.cif_file = cif_file
        if pets:self.init_pets(pets)
        if rock:
            self.rock = ut.load_pkl(rock)
            self.bloch=self.rock.load(self.frame)
            self.pets_opts,self.mode,self.nfigs = 'B','frames',self.rock.ts.size
            bloch = None
        if bloch:self.init_bloch(bloch)
        if multi:self.init_multi(multi)

        if not type(self.path)==str:
            self.path = os.path.dirname(self.cif_file)
        if not type(self.cif_file)==str:
            self.cif_file = mut._get_cif_file(self.path,cif_file)

        if bloch and pets :
            bloch_cif,pets_cif = os.path.basename(self.bloch.cif_file),os.path.basename(self.pets.cif_file)
            if not bloch_cif==pets_cif:
                raise Exception('bloch and pets seems to have different cif_files : bloch=%s, pets=%s'%(bloch_cif,pets_cif))

    def load_objects(self):
        if not type(self.bloch)==bl.Bloch:
            bloch_path = os.path.join(self.path,'bloch')
            if not os.path.exists(bloch_path):
                p = Popen("mkdir %s" %bloch_path,shell=True,stderr=PIPE,stdout=PIPE);p.wait()
            # print('creating Bloch')
            self.bloch = bl.Bloch(self.cif_file,path=bloch_path)#,solve=1)

        if not type(self.pets)==pt.Pets :
            if type(self.pets)==int:
                print(colors.green+'ignoring pets'+colors.black)
                self.pets = None
            else:
                pets_path = os.path.join(self.path,'pets')
                if not os.path.exists(pets_path):pets_path=self.path
                pts_files = glob.glob(os.path.join(pets_path,'*.pts'))
                if len(pts_files)>=1:
                    print(colors.green+'found .pts file. Loading %s ' %pts_files[0]+colors.black)
                    if not self.pets==-1:
                        self.init_pets(pts_files[0])
                else:
                    print(colors.red+'no pts files found'+colors.black)

        if not type(self.multi_path)==str :
            if type(self.multi_path)==int:
                print(colors.green+'ignoring multi'+colors.black)
                self.multi_path = None
            else:
                multi_path = os.path.join(self.path,'multi')
                if os.path.exists(multi_path):
                    self.multi_path = multi_path
                    msg ='''multislice folder found '''
                    print(colors.green+msg+colors.black)
                else:
                    print(colors.red+'no multislice folder'+colors.black)

        if self.multi_path:
            self.update_nfigs()
            self.multi = pp.load(self.multi_path,tag=self.tag+str(self.frame).zfill(3))
            if self.multi and not self.thick :
                self.thick = self.multi.thickness


        self.figpath = os.path.join(self.path,'figures')
        if not os.path.exists(self.figpath):
            p = Popen("mkdir %s" %self.figpath,shell=True,stderr=PIPE,stdout=PIPE);p.wait()

    def init_bloch(self,bloch):
        if type(bloch)   == bl.Bloch:self.bloch = bloch
        elif type(bloch) == str:self.bloch = load_bloch(file=bloch)
        else:raise Exception('bloch args need be NoneType, str or Bloch object ')
        if not type(self.path)     == str : self.path = self.bloch.path
        if not type(self.cif_file) == str : self.cif_file = self.bloch.cif_file
        self.u=self.bloch.u

    def init_multi(self,multi_path):
        if type(multi_path) in [str,int]:self.multi_path = multi_path

    def init_pets(self,pets):
        if type(pets)==pt.Pets : self.pets = pets
        elif type(pets)==str : self.pets = pt.Pets(pets,gen=1,cif_file=self.cif_file,dyn=self.dyn_opt)
        elif type(pets)==int : self.pets = pets;return
        else:raise Exception('pets args need be NoneType, str or Pets object ')
        if not type(self.path)     == str : self.path = self.pets.path
        if not type(self.cif_file) == str : self.cif_file = self.pets.cif_file
        self.nfigs    = self.pets.nFrames
        self.tifpath  = os.path.join(self.pets.path,'tiff')
        # self.fmt,self.load = mut.find_format(self.tifpath)
        # self.figs = np.sort(glob.glob(self.tifpath+'/*.%s' %self.fmt))#;print(self.figs)

    def open_config(self,config_file=''):
        ext = config_file.split('.')[-1]
        if not ext=='pkl':config_file=''
        elif not os.path.exists(config_file):
            config_file = os.path.join(self.path,config_file)

        if not os.path.exists(config_file):
            config_file = os.path.join(self.path,'')
            config_file = easygui.fileopenbox('open config','config',config_file,['*.pkl'])

        ext = config_file.split('.')[-1]
        if os.path.exists(config_file) and ext=='pkl' :
            print(colors.green+'Loading '+colors.yellow+config_file+colors.black)
            Viewer(config=config_file)

    def save_config(self,config_file=''):
        keys = ['path','cif_file','multi_path','tifpath','mode','tag','u',      #paths
        'cutoff','dcutoff','mag','xylims','Smax','Nmax','thick','dthick',       #generic
        'theta','phi','dtheta','dphi','dtheta_dphi','thicks',                   #bloch
        'frame','incrate','rot','rotS','drot','drotS','pets_opts','cmap',       #frames
        'show_hkl','show_i','show_Ig','show_Vg','show_Sw','Swtol','show_z','show_u','show_v',     #shows
        'save_bloch','hold_mode','cmap_beams',#'F','fopts',
        'multi_args','xyz_params','cond',
        ]
        if not config_file:
            config_file = os.path.join(self.path,'config.pkl')
            config_file = easygui.filesavebox('save config','save',config_file)

        if config_file:
            d_config = {k:self.__dict__[k] for k in keys}
            with open(config_file,'wb') as f:pickle.dump(d_config,f,pickle.HIGHEST_PROTOCOL)
            print(colors.yellow+config_file+colors.green+' saved'+colors.black)

    ######################################################################
    ##### call
    ######################################################################
    def __call__(self,event):
        ''' event callback function
        normal mode : call = call_update
        find mode   : call = shortcut_call
        '''
        self.call(event)

    def call_update(self, event):
        # print(event.key)
        key = event.key
        dfk = self.keymap.df_keys
        alias_match = dfk.loc[(dfk.alias==key) & ( (dfk['mode']==self.mode) | (dfk['mode']=='any'))]
        if any(alias_match.index):
            # print(alias_match)
            key = alias_match.iloc[0].key
        # print(key)
        self.get_keys(key)
        self.do_update(key)

    def find_shortcut(self):
        print(colors.green+'find shortcut mode :'+colors.black)
        msg = '''
        Type a shortcut to check its effect.
        Use escape to switch back to normal mode'''
        print(msg)
        self.call = self.shortcut_call

    def shortcut_call(self,event):
        key = event.key
        if key=='escape' :
            self.call=self.call_update
            print(colors.green+'Back to normal mode'+colors.black)
        else:
            self.keymap.shortcut_call(key)

    ######################################################################
    ##### update
    ######################################################################
    def update_u(self):
        if self.mode == 'rotate':
            theta,phi = np.deg2rad([self.theta,self.phi])
            ct,st,cp,sp = np.cos(theta),np.sin(theta),np.cos(phi),np.sin(phi)
            self.u = [st*cp,st*sp,ct]
        else:
            if self.tifpath:
                self.u = self.pets.get_beam(self.frame)
            elif self.multi:
                if 'u' in self.multi.__dict__.keys():
                    self.u = self.multi.u
                else:
                    print(colors.red+'warning : multislice does not contain u switching to rotate mode'+colors.black)
                    self.set_mode('rotate');return
                    self.u=[0,0,1]

        if self.multi_path:
            self.xyz_params['n']   = self.u
            self.xyz_params['xyz'] = self.get_xyz()


    def update_frame(self):
        if self.dsp_d['M']:
            loaded_multi = self.load_multi()
            if loaded_multi:
                state = self.multi.check_simu_state()       #;print(state)
                if state=='done':
                    self.multi.set_thicks()
                    self.multi.datpath=os.path.join(self.multi_path,'')
                    self.dthick = self.multi.dzs
                    self.update_thickness()
                else:
                    print(colors.blue+'multislice not done yet : '+colors.green+state+colors.black)
                    self.set_mode('rotate')
                    if not self.tifpath:return
            else:
                self.dsp_d['M'] = 0
                print(colors.red+'warning : no multislice simu found. Back to rotate mode'+colors.black)
                self.set_mode('rotate')
                if not self.tifpath:return

        if self.dsp_d['P'] and self.tifpath:
            self.rpl = self.pets.rpl.loc[self.pets.rpl.F==self.frame]
        if self.rock:
            self.bloch=self.rock.load(i=self.frame-1)
            self.u = self.bloch.u
        else:
            self.update(fsolve=1)#'B' in self.pets_opts)
        return 1

    def update_thickness(self):
        if self.dsp_d['M'] and self.mode=='frames':
            if self.thick<self.multi.thickness :
                self.iz,self.thick = self.multi.get_iz(self.thick,v=2)
                self.iz = int(self.iz)
            else:
                self.iz=None
                self.thick=self.multi.thickness
            qx,qy,I = self.multi.pattern(out=1,iz=self.iz,rot=self.rotS,v=1,**self.pattern_args)
            self.im=[qx,qy,I]
        if self.dsp_d['B'] or self.dsp_d['K']:
            self.bloch.set_thickness(self.thick)
            # print('update thickness',self.thick,self.bloch.thick)

    def update(self,fsolve=0):
        self.update_u()
        self.bloch.update_Nmax(self.Nmax)
        self.bloch.set_beam(u=self.u)                       #;print(self.u)
        self.bloch._set_excitation_errors(Smax=self.Smax)   #;print(self.Smax)
        self.bloch._set_Vg()
        self.bloch._set_kinematic()
        self.bloch.df_G['I']=0
        if fsolve:
            self.bloch._solve_Bloch()#opts='0v')
            self.bloch.set_thickness(thick=self.thick)
            if self.save_bloch:
                if self.mode=='frames':
                    bloch_file = os.path.join(self.path,'bloch','%s%s.pkl' %(self.tag,str(self.frame).zfill(4)))
                    self.bloch.save(bloch_file)
                else:self.bloch.save()
                self.save_bloch=0

    def set_mode(self,mode):
        self.mode = mode
        # self.show_im = {'frames':self.show_frames,'rotate':self.show_bloch}[self.mode]
        if self.mode=='frames':
            if self.update_frame():
                if self.dsp_d['M']:
                    self.dthick_r,self.dthick=self.dthick,self.multi.dzs
                    self.cutoff,self.mag = self.cutoff_r,self.mag_r
                    print('frames mode. Changing dthick=%d' %self.dthick)
        else:
            self.dthick=self.dthick_r
            self.cutoff_r,self.mag_r = self.cutoff,self.mag
            self.cutoff,self.mag = 50,500
            self.set_theta_phi_from_u(self.u)
            self.update(fsolve=1)

    ######################################################################
    ##### show
    ######################################################################
    def show_vals(self):
        if self.show_hkl : print(self.bloch.get_kin())
        if self.show_u   : print('[uvw]:');print(self.bloch.u)
        if self.show_v   : print('[uvw]_rec:');print(self.bloch.Kuvw0)
        if self.show_z   : print('zones:');print(self.bloch.get_zones())

        self.Icols=[] #,cond=[],[]
        if self.show_i  : self.Icols+=['I'] #;cond += ['(I>1e-4)']
        if self.show_Sw : self.Icols+=['Sw']#;cond += ['(Sw<1e-2)']
        if self.show_Ig : self.Icols+=['Ig']#;cond += ['(Ig>1e-2)']
        if self.show_Vg : self.Icols+=['Vg']#;cond += ['(Vga>1e-8)']
        # if any(cond):cond = self.cond+' & '.join(cond);print(cond)
        if self.cond and self.Icols:
            idx = self.bloch.get_beam(cond=self.cond)#;print(idx)
            cols = ['h','k','l']+self.Icols
            print(self.bloch.df_G.loc[idx,cols])

    def show_frames(self):
        title = ''
        sargs = {}
        cs = 'I'
        if not self.hold_mode:
            i,k = [(i,k) for i,(k,v) in enumerate(self.dsp_d.items()) if v][0]
            title += '%s, ' %list(EDdisp.legElts.keys())[i]
            cs = ['S','I'][k=='M']
        else:
            sargs={'alpha':0.75}
        if self.mode=='frames':
            title += 'frame %s, ' %self.frame_str()
        if self.multi_path and self.tag:
            title +='tag=%s, ' %(self.tag.replace('_',' '))
        if any([c in self.pets_opts for c in 'MKB']):
            title += 'thickness=%.1f$\AA$, ' %self.thick

        # sefl.refl = self.bloch.get_beam(cond=self.cond)#;print(hkl_idx)
        self.n,self.tol=5,1e-3
        self.refl = self.bloch.get_beam(cond=lambda dfG:bl_ut.strong_beams(dfG,tol=self.tol,n=self.n),index=False)
        # hkl_idx = self.refl
        print(self.refl)
        df = self.bloch.df_G
        df = df.drop(str((0,0,0)))
        EDdisp.show_frame(opts=self.pets_opts,mag=self.mag,rot=self.rot,
            df_pets=self.rpl,im_multi=self.im,df_bloch=df,hkl_idx=self.refl,
            ax=self.ax,title=title,xylims=self.xylims,single_mode=not self.hold_mode,
            cmap=self.cmap,cutoff=self.cutoff,sargs=sargs,cs=cs,**self.pargs)

    def show(self):
        self.show_vals()
        self.ax.cla()
        self.show_frames()
        self.fig.canvas.draw()

    ######################################################################
    ##### Multislice stuff
    ######################################################################
    def solve_multi(self):
        btns = ['Cancel','Change params','Gen xyz','Show xyz','Run']
        btn,tag = 'Start',self.tag
        # while (btn in btns[1:-2] or btn=='Start'):
        self.xyz_params['xyz']  = self.get_xyz()
        self.multi_args['data'] = self.get_xyz()
        self.multi_args['tail']  = self.tag+self.frame_str()
        msg = self.get_multi_params()
        btn = easygui.buttonbox(msg,'Multislice',btns)
        if   btn=='Change params' : self.set_multi()
        elif btn=='Gen xyz'       : self._gen_xyz(force=1);self.solve_multi()
        elif btn=='Show xyz'      : self.show_xyz()
        elif btn=='Run'           : self._run_multi()

        if not self.multi.tail=='_'+self.tag+self.frame_str():
            # print(self.multi.tail,self.tag+self.frame_str())
            self.update_frame()
            self.show()

    def set_multi(self):
        self.multi_path = os.path.join(self.path,'multi')
        if not os.path.exists(self.multi_path):
            p = Popen("mkdir %s" %self.multi_path,shell=True,stderr=PIPE,stdout=PIPE);p.wait()

        if not self.multi_args['ssh']:self.multi_args['ssh']=None
        if not self.tag:self.tag='None'
        mult_keys = ['NxNy','mulslice','slice_thick','i_slice','ssh','opt','Nhk','hk_pad']
        xyz_keys  = ['rep','lat_params','pad']
        fieldNames  = ['tag'] + mult_keys+xyz_keys
        fieldValues = [self.tag]
        fieldValues += [str(self.multi_args[k]) for k in mult_keys]
        fieldValues += [str(self.xyz_params[k]) for k in xyz_keys]
        vals = EDdisp.multenterbox('solve multislice','multi',fieldValues,fieldNames)
        if vals:
            self.xyz_params['lat_params'] = eval(vals['lat_params'])
            self.xyz_params['pad']        = eval(vals['pad'])
            self.xyz_params['rep']        = eval(vals['rep'])
            self.tag = vals['tag']
            if self.tag=='None':self.tag=''
            # self.frame = eval(vals['frame'])
            for f in mult_keys:
                if isinstance(self.multi_args[f],str):
                    self.multi_args[f] = vals[f]
                else:
                    self.multi_args[f] = eval(vals[f])
        self.solve_multi()

    def show_multi(self):
        self.show_xyz()
        msg = self.get_multi_params()
        cli = '''Multislice. Use :
        ctrl+M to solve
        ctrl+m to change params
        '''
        print(colors.green+cli+colors.black)
        print(msg)

    def show_xyz(self):
        self._gen_xyz();
        mut.show_grid(self.get_xyz(),opts=['xy','xz'],popts='hv',figs='f',title=os.path.basename(self.get_xyz()))
    def get_multi_params(self):
        df_multi = pd.DataFrame.from_dict(self.multi_args,orient='index',columns=['values'])
        df_xyz   = pd.DataFrame.from_dict(self.xyz_params,orient='index',columns=['values'])
        msg ='''
    xyz params :
        %s

    Multislice params :
        %s
        '''%(str(df_xyz),str(df_multi))
        return msg

    def _gen_xyz(self,force=0):
        xyz = self.get_xyz()
        if not os.path.exists(xyz) or force:
            print('generating xyz')
            # print(self.cif_file,xyz,self.u,
            #     self.lat_params,self.pad,self.rotS,'v')
            xyz_params = self.xyz_params.copy()
            if self.xyz_params['rep']:
                xyz_params.pop('lat_params');
                mut.gen_xyz(**xyz_params)
            else:
                xyz_params.pop('rep');
                mut.gen_xyz2(opts='v',**xyz_params)

    def open_multi(self):
        tails = self.show_tags(out=1)
        msg = '''Load multislice
        Available tags :
        %s
        ''' %tails
        tag = easygui.enterbox(msg,'load multislice',self.tag)
        if tag:
            if not tag==self.tag:
                self.tag=tag
                self.load_multi()
                self.update_frame()
    def load_multi(self):
            tag = self.tag+str(self.frame).zfill(3)
            multi = pp.load(self.multi_path,tag=tag)
            loaded_multi=False
            if multi:
                self.multi=multi
                self.multi.datpath = os.path.join(self.multi_path,'')
                self.multi.set_thicks() #; print("dz:",self.multi.dzs)
                self.multi.save()
                if 'xyz_params' in self.multi.__dict__.keys():
                    self.xyz_params = self.multi.xyz_params
                else:
                    print('warning : xyz_params not in multi')
                multi_args = list(self.multi_args.keys())
                for k in ['data','name','ssh','mulslice','opt']:multi_args.remove(k)
                self.multi_args['data'] = self.multi.data[0]
                for k in multi_args :
                    if k in self.multi.__dict__.keys():
                        self.multi_args[k]=self.multi.__dict__[k]
                    else:
                        print('warning : %s not in multi' %k)
                loaded_multi=True
            return loaded_multi

    def _run_multi(self):
        self._gen_xyz()
        # print('Solving multi : ', self.multi_args)
        self.multi = ms.Multislice(u=self.u,xyz_params=self.xyz_params,
            **self.multi_args)
        self.update_nfigs()

    def show_tags(self,out=0):
        if self.multi_path:
            pkls = glob.glob(os.path.join(self.multi_path,'*.pkl'))
            tails = ['_'.join(os.path.basename(pkl).split('_')[1:-1]) for pkl in pkls]
            tails = np.unique([tail[:-3] for tail in tails]) #;print(tails)
        if out:
            return tails
        else:
            print(colors.green+'available tails:'+colors.yellow,tails,colors.black)
    def get_xyz(self):
        return os.path.join(self.multi_path,self.tag+self.frame_str()+'.xyz')


    ######################################################################
    ##### Keys
    ######################################################################
    def do_update(self,key):
        if   key == self.keymap.dict_k['save_bloch']:self.update(fsolve=1)
        elif key in self.keymap.modes_keys : print('mode %s' %self.mode)
        elif key in self.keymap.frame_keys   and self.mode=='frames': print('frame %d' %self.frame)
        elif key in self.keymap.dframe_keys  and self.mode=='frames': print('increment rate : %d' %self.incrate)
        elif key in self.keymap.orient_keys  and self.mode=='rotate': print('theta=%.4f, phi=%.4f' %(self.theta,self.phi))
        elif key in self.keymap.dorient_keys and self.mode=='rotate': print('dtheta=%.1f, dphi=%.1f' %(self.dtheta,self.dphi))
        elif key in self.keymap.rot_keys     and self.mode=='frames': print('rot=%.1f, rotS=%.1f' %(self.rot,self.rotS))
        elif key in self.keymap.thick_keys      : print('thickness : %d' %self.thick);self.update_thickness()
        elif key in self.keymap.dthick_keys     : print('dthick=%d' %self.dthick)
        elif key in self.keymap.brightness_keys : print('brightness cutoff : %.1f' %self.cutoff)
        elif key in self.keymap.dbrightness_keys: print('dbrightness cutoff : %.1f' %self.dcutoff)
        elif key in self.keymap.rbrightness_keys: print('cutoff : %.1f, dcutoff : %.1f' %(self.cutoff,self.dcutoff))
        elif key in self.keymap.vals_keys       : self.show_vals()
        elif key in self.keymap.pets_keys       : print('frame options %s' %self.pets_opts)
        elif key == self.keymap.dict_k ['settings']:
            if self.mode=='rotate'   : self.update(fsolve=1)
            elif self.mode=='frames' : self.update_frame()
        #
        if self.mode=='rotate' :
            if key in self.keymap.orient_keys :
                self.update(fsolve=self.dsp_d['B'] and not self.bloch.solved )
        elif self.mode=='frames':
            if (key in self.keymap.frame_keys ) :
                self.update_frame()

        show_keys = key in self.keymap.show_keys
        show_keys = show_keys or (key in self.keymap.frame_keys  and self.mode=='frames')
        show_keys = show_keys or (key in self.keymap.orient_keys and self.mode=='rotate')
        if show_keys : self.show()#;print('showing')


    def get_keys(self,key):
        # print(key)
        dict_k = self.keymap.dict_k
        if   key == dict_k['settings']     : self.settings()
        elif key == dict_k['find_shortcut']: self.find_shortcut()
        elif key == dict_k['new_fig'] :
            config_file = os.path.join(self.path,'new_fig.pkl')
            self.save_config(config_file)
            Viewer(config=config_file)
            # v1.fig,v1.ax = dsp.stddisp()
            # cid = v1.fig.canvas.mpl_connect('key_press_event', v1)
        elif key == dict_k['help_cli']      : self.keymap.show_help_cli()
        elif key == dict_k['help_gui']      : self.keymap.show_help_gui()
        elif key == dict_k['help_simple']   : self.keymap.show_help_simple()
        elif key == dict_k['save_fig']      : dsp.saveFig(self.get_figname(),ax=self.ax)
        #config
        elif key == dict_k['save_config']   : self.save_config(os.path.join(self.path,'config.pkl'))
        elif key == dict_k['save_config_as']: self.save_config()
        elif key == dict_k['reload_config'] : self.open_config('config.pkl')
        elif key == dict_k['open_config']   : self.open_config()
        #bloch
        elif key == dict_k['save_bloch']    : self.save_bloch=1 #not self.save_bloch;print('%s saving bloch' %['not ',''][self.save_bloch])
        elif key == dict_k['solve_bloch']   : self.update(fsolve=1)
        #multi
        elif key == dict_k['open_multi']    : self.open_multi()
        elif key == dict_k['solve_multi']   : self.solve_multi()
        elif key == dict_k['set_multi']     : self.set_multi()
        elif key == dict_k['run_multi']     : self._run_multi()
        elif key == dict_k['show_multi']    : self.show_multi()
        elif key == dict_k['show_tags']     : self.show_tags()
        elif key == dict_k['log_info']      :
            if self.multi:print(self.multi.check_simu_state())
        #modes
        elif key == dict_k['frames_mode'] : self.set_mode('frames')
        elif key == dict_k['rotate_mode'] : self.set_mode('rotate')
        #in plane rotate
        if   key == dict_k['rot_up']   : self.rot=(self.rot+self.drot)%360
        elif key == dict_k['rot_down'] : self.rot=(self.rot-self.drot)%360
        elif key == dict_k['rotS_up']  : self.rotS=(self.rotS+self.drotS)%360
        elif key == dict_k['rotS_down']: self.rotS=(self.rotS-self.drotS)%360

        if self.mode=='rotate':
            # theta,phi
            if   key == dict_k['elev_up']   : self.theta=(self.theta+self.dtheta)%180
            elif key == dict_k['elev_down'] : self.theta=(self.theta-self.dtheta)%180
            elif key == dict_k['azim_up']   : self.phi=(self.phi+self.dphi)%360
            elif key == dict_k['azim_down'] : self.phi=(self.phi-self.dphi)%360
            # dtheta,dphi,dthick
            elif key == dict_k['delev_up']   : self.dtheta=min(self.dtheta+1,90)
            elif key == dict_k['delev_down'] : self.dtheta=max(self.dtheta-1,0)
            elif key == dict_k['dazim_up']   : self.dphi=min(self.dphi+1,90)
            elif key == dict_k['dazim_down'] : self.dphi=max(self.dphi-1,0)
        elif self.mode=='frames':
            # frame
            if   key == dict_k['frame_up']   : self.frame=min(self.frame+self.incrate,self.nfigs)
            elif key == dict_k['frame_down'] : self.frame=max(1,self.frame-self.incrate)
            # increment rate
            elif key == dict_k['inc_frame_up']  :self.incrate=min(self.incrate+1,100)
            elif key == dict_k['inc_frame_down']:self.incrate=max(1,self.incrate-1)

        #brightness
        if   key==dict_k['brightness up']     : self.cutoff=min(self.cutoff+self.dcutoff,1e5)
        elif key==dict_k['brightness down']   : self.cutoff=max(1e-5,self.cutoff-self.dcutoff)
        if   key==dict_k['dbrightness up']    : self.dcutoff*=2
        elif key==dict_k['dbrightness down']  : self.dcutoff/=2
        elif key==dict_k['brightness reset']  : self.cutoff  = 50
        elif key==dict_k['dbrightness reset'] : self.dcutoff = 10
        # thickness
        elif key == dict_k['thick_up']     : self.thick+=self.dthick
        elif key == dict_k['thick_down']   : self.thick=max(self.thick-self.dthick,self.dthick)
        elif key == dict_k['dthick_up']    : self.dthick=min(self.dthick+1,1000)
        elif key == dict_k['dthick_down']  : self.dthick=max(self.dthick-1,1)
        # vals
        elif key == dict_k['show_u']   : self.show_u   = not self.show_u  ;print('%sshowing [uvw]'     %['not ',''][self.show_u])
        elif key == dict_k['show_v']   : self.show_v   = not self.show_v  ;print('%sshowing [uvw]_rec' %['not ',''][self.show_v])
        elif key == dict_k['show_hkl'] : self.show_hkl = not self.show_hkl;print('%sshowing hkl'       %['not ',''][self.show_hkl])
        elif key == dict_k['show_I']   : self.show_i   = not self.show_i  ;print('%sshowing Istrong'   %['not ',''][self.show_i])
        elif key == dict_k['show_Ig']  : self.show_Ig  = not self.show_Ig ;print('%sshowing Ig_strong' %['not ',''][self.show_Ig])
        elif key == dict_k['show_Vg']  : self.show_Vg  = not self.show_Vg ;print('%sshowing Vg_strong' %['not ',''][self.show_Vg])
        elif key == dict_k['show_Sw']  : self.show_Sw  = not self.show_Sw ;print('%sshowing Sw_strong' %['not ',''][self.show_Sw])
        elif key == dict_k['show_z']   : self.show_z   = not self.show_z  ;print('%sshowing zones'     %['not ',''][self.show_z])
        elif key == dict_k['show_beams_vs_thickness']:
            if not self.bloch.solved : self.update(fsolve=1)
            # refl = self.bloch.get_beam(cond=lambda dfG:bl_rock.strong_beams(dfG,tol=0.01,n=5),index=False)
            # self.refl = self.bloch.get_beam(cond=lambda dfG:bl_ut.strong_beams(dfG,tol=0.01,n=10),index=False)

            self.bloch.show_beams_vs_thickness(self.thicks,cm=self.cmap_beams,refl=self.refl)
        elif key== dict_k['show_Exp'] and self.tifpath:
            self.pets.show_exp(frame=self.frame)
        # Display
        elif key == dict_k['hold']:
            self.hold_mode=not self.hold_mode
            if not self.hold_mode :
                k='K'
                if any(self.dsp_d.values()):k=self.pets_opts[0]
                self.dsp_d = self.dsp_d.fromkeys(self.dsp_d,False)
                self.dsp_d[k]=True
            print('hold mode ' + ['off','on'][self.hold_mode])

        do_up=False
        if key in self.keymap.opt_keys :
            self.opt_d[key] = not self.opt_d[key]
        elif key in self.keymap.dsp_keys:
            if self.hold_mode:
                self.dsp_d[key] = not self.dsp_d[key]
            else:
                self.dsp_d = self.dsp_d.fromkeys(self.dsp_d,False)
                self.dsp_d[key] = True
            do_up = do_up or (self.dsp_d['M'] and not 'M' in self.pets_opts and self.multi_path and not self.im)
            do_up = do_up or (self.dsp_d['P'] and not 'P' in self.pets_opts and self.tifpath )
            do_up = do_up or (self.dsp_d['B'] and not 'B' in self.pets_opts and not self.bloch.solved)
        self.set_pets_opts()
        if do_up:
            if self.mode=='frames' :self.update_frame()
            else:self.update(fsolve=1)

    ######################################################################
    ##### Misc
    ######################################################################
    def set_pets_opts(self):
        if not self.tifpath and self.dsp_d['P'] : self.dsp_d['P']=False;print(colors.red+'Processed data can not be displayed'+colors.black)
        if not self.multi   and self.dsp_d['M'] : self.dsp_d['M']=False;print(colors.red+'multislice can not be displayed'    +colors.black)
        if not self.hold_mode:
            if not any(self.dsp_d.values()):self.dsp_d['K']=True
        self.pets_opts =''.join([self.keymap.dsp_chars[c] for c,val in self.dsp_d.items() if val])
        self.pets_opts+=''.join([self.keymap.opt_chars[c] for c,val in self.opt_d.items() if val])
    def set_dsp_d(self):
        for k,v in self.keymap.dsp_chars.items():self.dsp_d[k] = v in self.pets_opts
        for k,v in self.keymap.opt_chars.items():self.opt_d[k] = v in self.pets_opts

    def set_theta_phi_from_u(self,u=None):
        if type(u) in [list,np.ndarray]:
            self.u = u/np.linalg.norm(u)
            x,y,z  = self.u
            self.theta = np.rad2deg(np.arccos(z))
            self.phi   = np.rad2deg(np.arctan2(y,x))
            print('u set to : ', str(self.u))
            print('theta=%.2f, phi=%.2f' %(self.theta,self.phi))

    def frame_str(self,pad=3):return '%s' %str(self.frame).zfill(pad)
    def get_figname(self):
        if self.mode=='frame':
            png_file = '_%s.png' %str(self.frame).zfill(3)
        else:
            png_file = '%s.png' %str(self.bloch.name)
        return os.path.join(self.figpath,png_file)
    def update_nfigs(self):
        if not self.tifpath and self.multi_path:
            self.nfigs = len(glob.glob(os.path.join(self.multi_path,'*.pkl')))

    def settings(self):
        fieldNames  =['Smax','Nmax','thick','dthick','cond',
            'xylims','mag','cutoff','Swtol','rot','rotS',
            'pets_opts','cmap','cmap_beams']
        if   self.mode=='rotate':
            self.dtheta_dphi=[self.dtheta,self.dphi]
            fieldNames+=['theta','phi','dtheta_dphi','thicks']
        elif self.mode=='frames':
            fieldNames+=['frame','incrate']
        if not self.cmap:self.cmap='None'
        fieldValues = [str(self.__dict__[f]) for f in fieldNames]
        dict_fv = EDdisp.multenterbox("Change settings","settings", fieldValues,fieldNames)
        if dict_fv:
            for f in fieldNames:
                if isinstance(self.__dict__[f],str):
                    self.__dict__[f] = dict_fv[f]
                else:
                    self.__dict__[f] = eval(dict_fv[f])

        if self.cmap == 'None':self.cmap=None
        self.set_dsp_d()
        #update dtheta,dphi
        if self.mode=='rotate':
            if type(self.dtheta_dphi) in [float,int]:
                self.dtheta_dphi=[self.dtheta_dphi]*2
            self.dtheta,self.dphi = self.dtheta_dphi
        self.m={'I':1e3,'Ig':1e2,'Vg':10,'Swl':self.Swtol}



'''Defines 2D lattices with some utilities'''
from utils import glob_colors as colors
from utils import displayStandards as dsp
import numpy as np

lat_types = ['square','rect','oblique','hex']

class Lattice2D:
    def __init__(self,lat_type='oblique',a=1,b=1,alpha=90):
        self.lat_type,self.a,self.b,self.alpha=lat_type,a,b,alpha
        self.params = {'a':a,'b':b,'alpha':alpha}
        self.lattice_vectors   =set_lattice_vec(lat_type,**self.params)
        self.reciprocal_vectors=get_reciprocal_vectors_2D(self.lattice_vectors)
        self.lat_vec = self.lattice_vectors
        self.rec_vec = self.reciprocal_vectors
        self.a1,self.a2 = self.lat_vec
        self.b1,self.b2 = self.rec_vec
        self.area=abs(np.cross(self.a1,self.a2))

    def get_vectors(self):return self.lattice_vectors
    def get_reciprocal_vectors(self):return self.reciprocal_vectors
    def get_lattice(self,nh,nk):
        ''' Get the miller indices and corresponding vectors
            - nh,nk : number of indices on each direction
        returns :
            - hk : Nx2 array of indices
            - R  : Nx2 array of vectors
        '''
        if type(nh) in [int,np.int64]:
            h,k = np.arange(nh),np.arange(nk)
            if nh<0:h = np.arange(nh,0)
            if nk<0:k = np.arange(nk,0)
        elif type(nh) in [np.ndarray,list]:
            h,k = nh,nk
        else:
            raise Exception('nh need be int or ndarray')
        h,k = np.meshgrid(h,k)
        hk  = np.vstack([h.flatten(),k.flatten()]).T
        R   = hk.dot(self.lattice_vectors) # all points of lattice
        return hk,R

    def plot_lattice(self,nh=3,nk=3,gridcolor=(0.5,)*3,hOpt=True,**kwargs) :
        '''Plot lattice grid :
        - nh,nk : number of cells along each direction
        - hOpt : additional lines for hexagonal lattice
        '''
        a1,a2=self.a1,self.a2
        hk,R = self.get_lattice(nh,nk)
        hs,ks = np.arange(nh+1),np.arange(nk+1)
        if nh<0:hs = np.arange(nh,1)
        if nk<0:ks = np.arange(nk,1)
        line_h = [h*a1 + np.array([[0,0],nk*a2]) for h in hs]
        line_k = [k*a2 + np.array([[0,0],nh*a1]) for k in ks]
        plts = [[xy[:,0],xy[:,1],gridcolor,''] for xy in line_h+line_k]
        if hOpt and self.lat_type=='hex':
            line_hex = [ np.vstack([r+a1,r+a2]) for r in R]
            plts += [[xy[:,0],xy[:,1],[gridcolor,'--'],''] for xy in line_hex]
        return dsp.stddisp(plts,**kwargs)

###########################################################################
# utils
###########################################################################
def get_reciprocal_vectors_2D(lat_vec):
    '''get reciprocal lattice vectors 2D
        - lat_vec : lattice vectors [a1;a2]
    returns :
        - rec_vec : lattice vectors [a1;a2]
    '''
    a1,a2   = lat_vec
    b1,b2   = reciprocal_lattice_2D(a1,a2)
    rec_vec = np.array([b1,b2])
    return rec_vec

def reciprocal_lattice_2D(a1,a2):
    Rot90 = np.array([[0,1],[-1,0]])
    b1 = Rot90.dot(a2)/(a1.dot(Rot90.dot(a2)))#*2*np.pi
    b2 = Rot90.dot(a1)/(a2.dot(Rot90.dot(a1)))#*2*np.pi
    return b1,b2

def set_lattice_vec(lat_type='',a=1,b=1,alpha=90,v=0) :
    ''' Computes lattice vectors from :
        - a,b,angle : lattice parameters
        - lat_type  : lattice types [hex,square,rect,oblique]
        - v : change orientation for oblique lattice
    '''
    if lat_type == 'square' :
        lattice_vec = np.array([[1,0],[0,1]])*a
    elif lat_type == 'rect' :
        lattice_vec = np.array([[a,0],[0,b]])
    elif lat_type == 'hex' :
        lattice_vec = np.array([[1,0],[np.cos(np.pi/3),np.sin(np.pi/3)]])*a
    elif lat_type == 'oblique' :
        angle=np.deg2rad(alpha)
        cp,sp = np.cos(angle),np.sin(angle)
        lattice_vec = np.array([[a,0],[b*cp,b*sp]])
        if v : lattice_vec = np.array([[a*cp,a*sp],[0,b]])
    else :
        msg ='''%s not a valid lattice type. choose either:
        %s''' %(lat_type,str(lat_types))
        raise Exception(msg)
    return lattice_vec


def params_from_lat_vec(lat_vec):
    a1,a2=lat_vec
    a = np.linalg.norm(a1)
    b = np.linalg.norm(a2)
    alpha = np.rad2deg(np.arccos(a1.dot(a2)/(a*b)))
    return {'a':a,'b':b,'alpha':alpha}

def get_miller(lattice_vec,nh,nk,ax=None):
    ''' get the miller indices and correspoding vectors in format :
            nh.nk x 2 '''
    n_h,n_k = np.arange(nh),np.arange(nk)
    miller = np.array(np.meshgrid(n_h,n_k)).T.reshape(((nh)*(nk),2))
    u_h = lattice_vec.T.dot(miller.T).T
    if ax :
        for u in u_h:
            ax.plot([0,u[0]],[0,u[1]],'s-g')
    return u_h,miller
